import osqp
from osqp.tests.utils import load_high_accuracy
import numpy as np
from scipy import sparse
import pytest
import numpy.testing as nptest


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    np.random.seed(4)

    self.n = 30
    self.m = 0
    P = sparse.diags(np.random.rand(self.n)) + 0.2 * sparse.eye(self.n)
    self.P = P.tocsc()
    self.q = np.random.randn(self.n)
    self.A = sparse.csc_matrix((self.m, self.n))
    self.l = np.array([])
    self.u = np.array([])
    self.opts = {
        'verbose': False,
        'eps_abs': 1e-08,
        'eps_rel': 1e-08,
        'polishing': False,
    }
    self.model = osqp.OSQP(algebra=algebra)
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, solver_type=solver_type, **self.opts)

    self.rtol = rtol
    self.atol = atol
    self.decimal_tol = decimal_tol

    return self


def test_unconstrained_problem(self):
    # Solve problem
    res = self.model.solve()

    # Assert close
    x_sol, _, obj_sol = load_high_accuracy('test_unconstrained_problem')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)



from types import SimpleNamespace
import osqp
from osqp.tests.utils import load_high_accuracy
import numpy as np
from scipy import sparse
import pytest
import numpy.testing as nptest


@pytest.fixture
def self(algebra, solver_type, atol, rtol, decimal_tol):
    self = SimpleNamespace()

    np.random.seed(1)

    self.n = 5
    self.m = 8
    p = 0.7

    Pt = sparse.random(self.n, self.n, density=p)
    Pt_new = Pt.copy()
    Pt_new.data += 0.1 * np.random.randn(Pt.nnz)

    self.P = (Pt.T.dot(Pt) + sparse.eye(self.n)).tocsc()
    self.P_new = (Pt_new.T.dot(Pt_new) + sparse.eye(self.n)).tocsc()
    self.P_triu = sparse.triu(self.P)
    self.P_triu_new = sparse.triu(self.P_new)
    self.q = np.random.randn(self.n)
    self.A = sparse.random(self.m, self.n, density=p, format='csc')
    self.A_new = self.A.copy()
    self.A_new.data += np.random.randn(self.A_new.nnz)
    self.l = np.zeros(self.m)
    self.u = 30 + np.random.randn(self.m)
    self.opts = {'eps_abs': 1e-08, 'eps_rel': 1e-08, 'verbose': False}
    self.model = osqp.OSQP(algebra=algebra)
    self.model.setup(P=self.P, q=self.q, A=self.A, l=self.l, u=self.u, solver_type=solver_type, **self.opts)

    self.rtol = rtol
    self.atol = atol
    self.decimal_tol = decimal_tol

    return self


def test_solve(self):
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_solve')
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P(self):
    # Update matrix P
    Px = self.P_triu_new.data
    Px_idx = np.arange(self.P_triu_new.nnz)
    self.model.update(Px=Px, Px_idx=Px_idx)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P')

    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P_allind(self):
    # Update matrix P
    Px = self.P_triu_new.data
    self.model.update(Px=Px)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P_allind')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_A(self):
    # Update matrix A
    Ax = self.A_new.data
    Ax_idx = np.arange(self.A_new.nnz)
    self.model.update(Ax=Ax, Ax_idx=Ax_idx)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_A')

    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_A_allind(self):
    # Update matrix A
    Ax = self.A_new.data
    self.model.update(Ax=Ax)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_A_allind')
    # Assert close
    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P_A_indP_indA(self):
    # Update matrices P and A
    Px = self.P_triu_new.data
    Px_idx = np.arange(self.P_triu_new.nnz)
    Ax = self.A_new.data
    Ax_idx = np.arange(self.A_new.nnz)
    self.model.update(Px=Px, Px_idx=Px_idx, Ax=Ax, Ax_idx=Ax_idx)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P_A_indP_indA')

    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P_A_indP(self):
    # Update matrices P and A
    Px = self.P_triu_new.data
    Px_idx = np.arange(self.P_triu_new.nnz)
    Ax = self.A_new.data
    self.model.update(Px=Px, Px_idx=Px_idx, Ax=Ax)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P_A_indP')

    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P_A_indA(self):
    # Update matrices P and A
    Px = self.P_triu_new.data
    Ax = self.A_new.data
    Ax_idx = np.arange(self.A_new.nnz)
    self.model.update(Px=Px, Ax=Ax, Ax_idx=Ax_idx)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P_A_indA')

    if self.model.algebra != 'cuda':  # pytest-todo
        nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
        nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)


def test_update_P_A_allind(self):
    # Update matrices P and A
    Px = self.P_triu_new.data
    Ax = self.A_new.data
    self.model.update(Px=Px, Ax=Ax)
    res = self.model.solve()

    x_sol, y_sol, obj_sol = load_high_accuracy('test_update_P_A_allind')

    nptest.assert_allclose(res.x, x_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_allclose(res.y, y_sol, rtol=self.rtol, atol=self.atol)
    nptest.assert_almost_equal(res.info.obj_val, obj_sol, decimal=self.decimal_tol)



import numpy.random as npr
import numpy as np
import torch
import numpy.testing as npt
import scipy.sparse as spa
from scipy.optimize import approx_fprime
import pytest

import osqp
from osqp.nn.torch import OSQP

ATOL = 1e-2
RTOL = 1e-4
EPS = 1e-5

cuda = False
verbose = True


# Note (02/13/24)
# Some versions of Python/torch/numpy cannot coexist on certain platforms.
# This is a problem seen with numpy>=2.
# Support is gradually being added. Rather than keep track of which versions
# are supported and on what platforms (which is likely to change frequently),
# we do an early check that is seen to raise RuntimeError in these cases,
# and skip testing this module entirely.
try:
    torch.ones(1).cpu().numpy()
except RuntimeError:
    pytest.skip('torch/numpy mutual incompatibility', allow_module_level=True)


def get_grads(
    n_batch=1,
    n=10,
    m=3,
    P_scale=1.0,
    A_scale=1.0,
    u_scale=1.0,
    l_scale=1.0,
    algebra=None,
    solver_type=None,
):
    assert n_batch == 1
    npr.seed(1)
    L = np.random.randn(n, n)
    P = spa.csc_matrix(P_scale * L.dot(L.T))
    x_0 = npr.randn(n)
    s_0 = npr.rand(m)
    A = spa.csc_matrix(A_scale * npr.randn(m, n))
    u = A.dot(x_0) + A_scale * s_0
    l = -10 * A_scale * npr.rand(m)
    q = npr.randn(n)
    true_x = npr.randn(n)

    P, q, A, l, u, true_x = [x.astype(np.float64) for x in [P, q, A, l, u, true_x]]

    grads = get_grads_torch(P, q, A, l, u, true_x, algebra, solver_type)
    return [P, q, A, l, u, true_x], grads


def get_grads_torch(P, q, A, l, u, true_x, algebra, solver_type):
    P_idx = P.nonzero()
    P_shape = P.shape
    A_idx = A.nonzero()
    A_shape = A.shape

    P_torch, q_torch, A_torch, l_torch, u_torch, true_x_torch = [
        torch.DoubleTensor(x) if len(x) > 0 else torch.DoubleTensor() for x in [P.data, q, A.data, l, u, true_x]
    ]
    if cuda:
        P_torch, q_torch, A_torch, l_torch, u_torch, true_x_torch = [
            x.cuda() for x in [P.data, q, A.data, l, u, true_x]
        ]

    for x in [P_torch, q_torch, A_torch, l_torch, u_torch]:
        x.requires_grad = True

    x_hats = OSQP(
        P_idx,
        P_shape,
        A_idx,
        A_shape,
        algebra=algebra,
        solver_type=solver_type,
    )(P_torch, q_torch, A_torch, l_torch, u_torch)

    dl_dxhat = x_hats.data - true_x_torch
    x_hats.backward(dl_dxhat)

    grads = [x.grad.data.squeeze(0).cpu().numpy() for x in [P_torch, q_torch, A_torch, l_torch, u_torch]]
    return grads


def test_dl_dq(algebra, solver_type, atol, rtol, decimal_tol):
    n, m = 5, 5

    model = osqp.OSQP(algebra=algebra)
    if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
        pytest.skip('No derivatives capability')

    [P, q, A, l, u, true_x], [dP, dq, dA, dl, du] = get_grads(
        n=n,
        m=m,
        P_scale=100.0,
        A_scale=100.0,
        algebra=algebra,
        solver_type=solver_type,
    )

    def f(q):
        model.setup(P, q, A, l, u, solver_type=solver_type, verbose=False)
        res = model.solve()
        x_hat = res.x

        return 0.5 * np.sum(np.square(x_hat - true_x))

    dq_fd = approx_fprime(q, f, epsilon=EPS)
    if verbose:
        print('dq_fd: ', np.round(dq_fd, decimals=4))
        print('dq: ', np.round(dq, decimals=4))
    npt.assert_allclose(dq_fd, dq, rtol=RTOL, atol=ATOL)


def test_dl_dP(algebra, solver_type, atol, rtol, decimal_tol):
    n, m = 5, 5

    model = osqp.OSQP(algebra=algebra)
    if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
        pytest.skip('No derivatives capability')

    [P, q, A, l, u, true_x], [dP, dq, dA, dl, du] = get_grads(
        n=n,
        m=m,
        P_scale=100.0,
        A_scale=100.0,
        algebra=algebra,
        solver_type=solver_type,
    )

    def f(P):
        P = P.reshape(n, n)
        P = spa.csc_matrix(P)
        model.setup(P, q, A, l, u, solver_type=solver_type, verbose=False)
        res = model.solve()
        x_hat = res.x

        return 0.5 * np.sum(np.square(x_hat - true_x))

    dP_fd = approx_fprime(P.toarray().flatten(), f, epsilon=EPS)
    if verbose:
        print('dP_fd: ', np.round(dP_fd, decimals=4))
        print('dP: ', np.round(dP, decimals=4))
    npt.assert_allclose(dP_fd, dP, rtol=RTOL, atol=ATOL)


def test_dl_dA(algebra, solver_type, atol, rtol, decimal_tol):
    n, m = 5, 5

    model = osqp.OSQP(algebra=algebra)
    if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
        pytest.skip('No derivatives capability')

    [P, q, A, l, u, true_x], [dP, dq, dA, dl, du] = get_grads(
        n=n,
        m=m,
        P_scale=100.0,
        A_scale=100.0,
        algebra=algebra,
        solver_type=solver_type,
    )

    def f(A):
        A = A.reshape((m, n))
        A = spa.csc_matrix(A)
        model.setup(P, q, A, l, u, solver_type=solver_type, verbose=False)
        res = model.solve()
        x_hat = res.x

        return 0.5 * np.sum(np.square(x_hat - true_x))

    dA_fd = approx_fprime(A.toarray().flatten(), f, epsilon=EPS)
    if verbose:
        print('dA_fd: ', np.round(dA_fd, decimals=4))
        print('dA: ', np.round(dA, decimals=4))
    npt.assert_allclose(dA_fd, dA, rtol=RTOL, atol=ATOL)


def test_dl_dl(algebra, solver_type, atol, rtol, decimal_tol):
    n, m = 5, 5

    model = osqp.OSQP(algebra=algebra)
    if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
        pytest.skip('No derivatives capability')

    [P, q, A, l, u, true_x], [dP, dq, dA, dl, du] = get_grads(
        n=n,
        m=m,
        P_scale=100.0,
        A_scale=100.0,
        algebra=algebra,
        solver_type=solver_type,
    )

    def f(l):
        model.setup(P, q, A, l, u, solver_type=solver_type, verbose=False)
        res = model.solve()
        x_hat = res.x

        return 0.5 * np.sum(np.square(x_hat - true_x))

    dl_fd = approx_fprime(l, f, epsilon=EPS)
    if verbose:
        print('dl_fd: ', np.round(dl_fd, decimals=4))
        print('dl: ', np.round(dl, decimals=4))
    npt.assert_allclose(dl_fd, dl, rtol=RTOL, atol=ATOL)


def test_dl_du(algebra, solver_type, atol, rtol, decimal_tol):
    n, m = 5, 5

    model = osqp.OSQP(algebra=algebra)
    if not model.has_capability('OSQP_CAPABILITY_DERIVATIVES'):
        pytest.skip('No derivatives capability')

    [P, q, A, l, u, true_x], [dP, dq, dA, dl, du] = get_grads(
        n=n,
        m=m,
        P_scale=100.0,
        A_scale=100.0,
        algebra=algebra,
        solver_type=solver_type,
    )

    def f(u):
        model.setup(P, q, A, l, u, solver_type=solver_type, verbose=False)
        res = model.solve()
        x_hat = res.x

        return 0.5 * np.sum(np.square(x_hat - true_x))

    du_fd = approx_fprime(u, f, epsilon=EPS)
    if verbose:
        print('du_fd: ', np.round(du_fd, decimals=4))
        print('du: ', np.round(du, decimals=4))
    npt.assert_allclose(du_fd, du, rtol=RTOL, atol=ATOL)
